package net.knarcraft.ffmpegconverter.container;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.List;

/**
 * A class for generating and storing a ffmpeg command
 */
public class FFMpegCommand {

    private final @NotNull String executable;
    private final @NotNull List<String> globalOptions;
    private final @NotNull List<String> inputFileOptions;
    private final @NotNull List<String> inputFiles;
    private final @NotNull List<String> outputFileOptions;
    private @Nullable String outputVideoCodec;
    private @NotNull String outputFile;

    /**
     * Instantiates a new FFMPEG command
     *
     * @param executable <p>The FFMPEG/FFPROBE executable to run</p>
     */
    public FFMpegCommand(@NotNull String executable) {
        this.executable = executable;
        this.globalOptions = new ArrayList<>();
        this.inputFileOptions = new ArrayList<>();
        this.inputFiles = new ArrayList<>();
        this.outputFileOptions = new ArrayList<>();
        this.outputFile = "";
    }

    /**
     * Adds a global ffmpeg option to this command
     *
     * @param argument <p>The option(s) to add</p>
     */
    public void addGlobalOption(@NotNull String... argument) {
        this.globalOptions.addAll(List.of(argument));
    }

    /**
     * Adds an input file option to this command
     *
     * @param argument <p>The input file option(s) to add</p>
     */
    public void addInputFileOption(@NotNull String... argument) {
        this.inputFileOptions.addAll(List.of(argument));
    }

    /**
     * Adds an input file to this command
     *
     * <p>Note that this adds the "-i", so don't add that yourself!</p>
     *
     * @param argument <p>The input file(s) to add</p>
     */
    public void addInputFile(@NotNull String... argument) {
        for (String fileName : argument) {
            if (fileName.isEmpty()) {
                continue;
            }
            this.inputFiles.add(fileName);
        }
    }

    /**
     * Gets the input files currently added to this command
     *
     * @return <p>The input files</p>
     */
    @NotNull
    public List<String> getInputFiles() {
        return new ArrayList<>(this.inputFiles);
    }

    /**
     * Adds an output file option to this command
     *
     * @param argument <p>The output file option(s) to add</p>
     */
    public void addOutputFileOption(@NotNull String... argument) {
        this.outputFileOptions.addAll(List.of(argument));

        // Detect when the output video codec is set
        for (int i = 0; i < argument.length; i++) {
            String item = argument[i].trim().toLowerCase();
            if (i < argument.length - 1 && item.equals("-vcodec") || item.equals("-codec:v") || item.equals("-c:v")) {
                this.outputVideoCodec = argument[i + 1];
            }
        }
    }

    /**
     * Sets the output file for this command
     *
     * @param argument <p>The path to the output file</p>
     */
    public void setOutputFile(@NotNull String argument) {
        this.outputFile = argument;
    }

    /**
     * Gets the output video codec set in this command
     *
     * @return <p>The output video codec, or null if not set</p>
     */
    @Nullable
    public String getOutputVideoCodec() {
        return this.outputVideoCodec;
    }

    /**
     * Gets the result of combining all the given input
     *
     * @return <p>The generated FFMPEG command</p>
     */
    @NotNull
    public String[] getResult() {
        List<String> result = new ArrayList<>();
        result.add(executable);
        result.addAll(globalOptions);
        result.addAll(inputFileOptions);
        for (String inputFile : inputFiles) {
            result.add("-i");
            result.add(inputFile);
        }
        result.addAll(outputFileOptions);
        if (!outputFile.isEmpty()) {
            result.add(outputFile);
        }
        return result.toArray(new String[0]);
    }

}
