package net.knarcraft.ffmpegconverter.converter.sorter;

import net.knarcraft.ffmpegconverter.property.MinimalSubtitlePreference;
import net.knarcraft.ffmpegconverter.streams.SubtitleStream;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;

/**
 * A sorter for sorting/filtering subtitles by a minimal subtitle preference
 */
public class MinimalSubtitleSorter extends AbstractSorter<SubtitleStream> {

    private final MinimalSubtitlePreference minimalSubtitlePreference;

    /**
     * Instantiates a new minimal subtitle preference sorter
     *
     * @param minimalSubtitlePreference <p>The minimal subtitle preference sort/filter by</p>
     */
    public MinimalSubtitleSorter(@NotNull MinimalSubtitlePreference minimalSubtitlePreference) {
        this.minimalSubtitlePreference = minimalSubtitlePreference;
    }

    @Override
    public @NotNull List<SubtitleStream> sort(@NotNull List<SubtitleStream> input) {
        // Split all subtitles into full and minimal
        List<SubtitleStream> fullSubtitles = new ArrayList<>();
        List<SubtitleStream> minimalSubtitles = new ArrayList<>();
        for (SubtitleStream subtitleStream : input) {
            if (subtitleStream.isFullSubtitle()) {
                fullSubtitles.add(subtitleStream);
            } else {
                minimalSubtitles.add(subtitleStream);
            }
        }

        // Sort/filter subtitles based on full and minimal
        switch (this.minimalSubtitlePreference) {
            case REJECT -> {
                // Only return full subtitles
                return fullSubtitles;
            }
            case REQUIRE -> {
                // Only return minimal subtitles
                return minimalSubtitles;
            }
            case NO_PREFERENCE -> {
                // Don't change order
                return input;
            }
            case PREFER -> {
                // Sort minimal subtitles first, and full subtitles last
                minimalSubtitles.addAll(fullSubtitles);
                return minimalSubtitles;
            }
            case AVOID -> {
                // Sort full subtitles first, and minimal subtitles last
                fullSubtitles.addAll(minimalSubtitles);
                return fullSubtitles;
            }
            default -> throw new IllegalStateException("Unknown enum value encountered");
        }
    }

}
